#include "main.h"

#include <EGL/egl.h>
#include <GLES2/gl2.h>
#include <GLES2/gl2ext.h>

//////////////////////////////////////////////////////////////////////////////
// static inlines...
#include "mat.c"
//////////////////////////////////////////////////////////////////////////////

static float matr_orth[16];

typedef struct
{
   int prog;
   struct {
      unsigned int vert;
      unsigned int tex;
   } attrib;
   struct {
      unsigned int mvp;
      unsigned int texture;
      unsigned int res;
   } uniform;
} Prog;

typedef struct
{
   unsigned int tex, fb;
   int w, h;
} Buf;

Buf *
buf_tex_new(unsigned char *pix, int w, int h)
{
   Buf *buf = calloc(1, sizeof(*buf));

   buf->w = w;
   buf->h = h;
   glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
   glGenTextures(1, &(buf->tex));
   glBindTexture(GL_TEXTURE_2D, buf->tex);
   glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, w, h, 0, GL_RGBA, GL_UNSIGNED_BYTE, pix);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
   return buf;
}

Buf *
buf_tex_file(char *file, int *w, int *h)
{
   unsigned char *pix = png_load(file, w, h);
   if (!pix) return NULL;
   Buf *buf = buf_tex_new(pix, *w, *h);
   free(pix);
   return buf;
}

Buf *
buf_fbo_new(int w, int h)
{
   Buf *buf = calloc(1, sizeof(*buf));
   int fnum;

   buf->w = w;
   buf->h = h;
   glGetIntegerv(GL_FRAMEBUFFER_BINDING, &fnum); // get fb out

   glGenFramebuffers(1, &(buf->fb));
   if (glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE)
     {
        printf("FBO: create %ix%i failed\n", w, h);
        free(buf);
        return NULL;
     }
   glBindFramebuffer(GL_FRAMEBUFFER, buf->fb);

   glGenTextures(1, &(buf->tex));
   glBindTexture(GL_TEXTURE_2D, buf->tex);
   glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, buf->w, buf->h, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
   glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, buf->tex, 0);

   glBindFramebuffer(GL_FRAMEBUFFER, fnum); // restore fb out
   return buf;
}

Buf *
buf_back_new(int w, int h)
{
   Buf *buf = calloc(1, sizeof(*buf));
   buf->w = w;
   buf->h = h;
   return buf;
}

void
buf_dest(Buf *buf)
{
   glViewport(0, 0, buf->w, buf->h);
   if (buf->fb)
     {
        mat_ortho(matr_orth, buf->w, buf->h, 1/* vflip */);
        glBindFramebuffer(GL_FRAMEBUFFER, buf->fb);
     }
   else
     {
        mat_ortho(matr_orth, buf->w, buf->h, 0/* vflip */);
        glBindFramebuffer(GL_FRAMEBUFFER, 0);
     }
}

void
buf_src(Buf *buf)
{
   if (buf->tex) glBindTexture(GL_TEXTURE_2D, buf->tex);
   else glBindTexture(GL_TEXTURE_2D, 0);
}

void
img_draw(Buf *src, Prog *prog, int x, int y, int w, int h)
{
   float verts[6 * 2] = {
      x    , y    ,   x + w, y,       x    , y + h,
      x    , y + h,   x + w, y,       x + w, y + h
   };
   float texuv[6 * 2] = {
      0, 0,   1, 0,   0, 1,
      0, 1,   1, 0,   1, 1
   };

   glDisable(GL_BLEND);
   glEnable(GL_BLEND);
   glActiveTexture(GL_TEXTURE0);

   buf_src(src);

   glUseProgram(prog->prog);

   glUniformMatrix4fv(prog->uniform.mvp, 1, GL_FALSE, matr_orth);
   glUniform2f(prog->uniform.res, src->w, src->h);
   glUniform1i(prog->uniform.texture, 0);

   glEnableVertexAttribArray(prog->attrib.vert); // verts
   glVertexAttribPointer(prog->attrib.vert, 2, GL_FLOAT, GL_TRUE, 0, verts);

   glEnableVertexAttribArray(prog->attrib.tex); // texuv
   glVertexAttribPointer(prog->attrib.tex, 2, GL_FLOAT, GL_TRUE, 0, texuv);

   glDrawArrays(GL_TRIANGLES, 0, 6);
}

static void
_prog_link(Prog *prog)
{
   prog->uniform.mvp     = glGetUniformLocation(prog->prog, "_unif_mvp");
   prog->uniform.texture = glGetUniformLocation(prog->prog, "_unif_texture");
   prog->uniform.res     = glGetUniformLocation(prog->prog, "_unif_res");

   prog->attrib.vert     = glGetAttribLocation(prog->prog, "_attr_vertex");
   prog->attrib.tex      = glGetAttribLocation(prog->prog, "_attr_texuv");
}

Prog *
prog_new(const char *shd_vert, const char *shd_frag)
{
   Prog *prog = calloc(1, sizeof(*prog));

   prog->prog = shad_prog_new(shd_vert, shd_frag);
   _prog_link(prog);
   return prog;
}

static char *
_file_load(const char  *file)
{
   char *str = NULL;
   FILE *f = fopen(file, "r");
   long size;
   if (!f) return NULL;
   fseek(f, 0, SEEK_END);
   size = ftell(f);
   rewind(f);
   str = malloc(size + 1);
   fread(str, size, 1, f);
   str[size] = 0;
   fclose(f);
   return str;
}

Prog *
prog_file_new(const char *shd_vert_file, const char *shd_frag_file)
{
   Prog *prog = calloc(1, sizeof(*prog));
   char *shd_vert = _file_load(shd_vert_file);
   char *shd_frag = _file_load(shd_frag_file);

   if ((!shd_vert) || (!shd_frag))
     {
        if (!shd_vert) printf("Cannot load vertex shader: %s\n", shd_vert_file);
        if (!shd_frag) printf("Cannot load fragment shader: %s\n", shd_frag_file);
        free(prog);
        prog = NULL;
        goto err;
     }
   prog->prog = shad_prog_new(shd_vert, shd_frag);
   _prog_link(prog);
err:
   free(shd_frag);
   free(shd_vert);
   return prog;
}

static Buf *buf_back = NULL;
static Buf *buf_bg = NULL;
static Buf *buf_img = NULL;
static Buf *buf_tmp = NULL;
static Buf *buf_tmp2 = NULL;

static Prog *prog_plain = NULL;
static Prog *prog_blurh3 = NULL;
static Prog *prog_blurv3 = NULL;
static Prog *prog_blurh5 = NULL;
static Prog *prog_blurv5 = NULL;
static Prog *prog_blurh7 = NULL;
static Prog *prog_blurv7 = NULL;

static int bgw = 0, bgh = 0;
static int imw = 0, imh = 0;

void
draw_init(int w, int h)
{
   buf_bg = buf_tex_file("bg.png", &bgw, &bgh);
   buf_img = buf_tex_file("image.png", &imw, &imh);
   buf_back = buf_back_new(w, h);
   buf_tmp = buf_fbo_new(imw / 2, imh / 2);
   buf_tmp2 = buf_fbo_new(imw / 2, imh / 2);

   prog_plain  = prog_file_new("shad_plain.vert.glsl", "shad_plain.frag.glsl");
   prog_blurh3 = prog_file_new("shad_blurh3.vert.glsl", "shad_blurh3.frag.glsl");
   prog_blurv3 = prog_file_new("shad_blurv3.vert.glsl", "shad_blurv3.frag.glsl");
   prog_blurh5 = prog_file_new("shad_blurh5.vert.glsl", "shad_blurh5.frag.glsl");
   prog_blurv5 = prog_file_new("shad_blurv5.vert.glsl", "shad_blurv5.frag.glsl");
   prog_blurh7 = prog_file_new("shad_blurh7.vert.glsl", "shad_blurh7.frag.glsl");
   prog_blurv7 = prog_file_new("shad_blurv7.vert.glsl", "shad_blurv7.frag.glsl");

   glDisable(GL_DEPTH_TEST);
}

void
draw(int w, int h)
{
   int i;
   unsigned char pixdata[4];
   double t0, t;
   int iter = 2000;
   const char *s;

   buf_dest(buf_back);

   // clear
   glClearColor(0.15, 0.25, 0.35, 1.0);
   glClear(GL_COLOR_BUFFER_BIT);
   // place bg image centered
   buf_dest(buf_tmp);
   img_draw(buf_bg, prog_plain, (w - bgw) / 2, (h - bgh) / 2, bgw, bgh);

   glReadPixels(w / 2, h / 2, 1, 1, GL_RGBA, GL_UNSIGNED_BYTE, pixdata);
   s = getenv("ITER");
   if (s) iter = atoi(s);
   t0 = get_time();
   for (i = 0; i < iter; i++)
     {
        // draw img to fbo at 1/2 res
        buf_dest(buf_tmp);
        img_draw(buf_img, prog_plain, 0, 0, imw / 2, imh / 2);

        // draw img to fbo2 with horiz blur
        buf_dest(buf_tmp2);
        img_draw(buf_tmp, prog_blurh7, 0, 0, imw / 2, imh / 2);

        // draw fbo2 to backbuffer with vblur and upscale 2x
        buf_dest(buf_back);
        img_draw(buf_tmp2, prog_blurv7, rand() % 0xf, rand() & 0xf, imw, imh);
     }
   glReadPixels(w / 2, h / 2, 1, 1, GL_RGBA, GL_UNSIGNED_BYTE, pixdata);
   t = get_time();
   printf("TIME: %1.8f\n", t - t0);
}
